/* This program solves the 0-1 Knapsack problem using Branch and Bound
 * For information about the Knapsack problem, see
 * "Knapsack Problems. Algorithms and Computer Implementations." by
 * Silvano Martello and Paolo Toth. John Wiley and Sons.
 * For information on Branch and Bound use the book of Horowitz and Sahni
 * "Fundamentals of Computers Algorithms" (Pitman)
 * This is approach focuses in the aspects of general
 * parallel branch and bound and does not in the concrete
 * features of knapsack problems.
 * Author: Casiano Rodriguez Leon. Email: casiano@ull.es
*/
#undef FORK95

#ifdef FORK95
#include <fork.h>
#include <io.h>
#else
#include <stdio.h>
#endif
#include <stdlib.h>

#define MAX 4096
#define MAXHEAPSIZE 16
#define STRIDE 1
#define  INFINITY 8192

typedef int number;
typedef int boolean;

typedef struct {
	int k, C, P;
} problemt;

#define ns "%d\n"

number N, M;
number w[MAX], p[MAX];
number bestSol = -INFINITY;

problemt problems[MAXHEAPSIZE];
number head = 0, tail = 0;

FILE *in;

void readKnap(void) {
  number i;

  if ((in = fopen("knap.dat","r+")) == NULL) 
    fprintf(stderr, "Cannot open input file.\n");
  fscanf(in,ns,&N);
  fscanf(in,ns,&M);
  printf("N = "); printf(ns,N);
  printf("M = "); printf(ns,M);
  for(i=0;i<N;i++) {
    fscanf(in,ns,&w[i]);
    fscanf(in,ns,&p[i]);
    if (i%STRIDE == 0) {
      printf("w = "); printf(ns,w[i]);
      printf("p = "); printf(ns,p[i]);
    }  
  }
  w[N] = INFINITY; p[N] = 0; /* sentinel object */
  fclose(in);
}

void lowerUpper(number k, number C, number P,number *L, number *U) {
  /* naive approach */
  number i, weig, prof;

  if (C < 0) {
	 *L = -INFINITY; *U = -INFINITY;
  }
  else {
	 for(i=k, weig = 0, prof = P; weig <= C; i++) {
		weig += w[i]; prof += p[i];
	 }
	 /* invariant:  weig > C */
	 i--;
	 weig -= w[i];
	 prof -= p[i];
	 *L = prof;  /* see Martello and Toth pag 16. Th 2.1 */
	 *U = prof+(p[i]*(C-weig))/w[i];
  }
}

void getProblem(number * k, number * C, number * P) {

  *k = problems[head].k;
  *C = problems[head].C;
  *P = problems[head].P;

  head = (head+1)% MAXHEAPSIZE;
}

void insertProblem(number k, number C, number P) {
  number next;

  next = (tail+1)%MAXHEAPSIZE;

  if (next == head) {
	 printf("\nError. The heap of problems is full.");
	 printf("\nIncrease the value of MAXHEAPSIZE. Current value is (%d)\n",MAXHEAPSIZE);
	 exit(1);
  }
  problems[tail].k = k;
  problems[tail].C = C;
  problems[tail].P = P;
  tail = next;
}

int numProblems(void) {
  return (((tail-head)+MAXHEAPSIZE)%MAXHEAPSIZE);
}

boolean inner() {
  return (k < N);
}
void branchAndBound(void) {
  number k, C, P, L, U;

  while (numProblems() > 0) {
	 getProblem(&k, &C, &P);
	 if  inner() {
		lowerUpper(k, C, P, &L, &U);
		if (bestSol < L) bestSol = L;
		/* L <= bestSol <= U */
		if (bestSol < U) {
		  int next;

		  next = k+1;
		  insertProblem(next,C-w[k], P+p[k]);
		  insertProblem(next,C, P);
		}
	 } /* end k < N */
  } /* end while */
}

int main(void) {

  readKnap();
  insertProblem(0 /* next object */,
					 M /* remaining capacity */,
					 0 /* profit */
					);
  branchAndBound();
  printf("\nsol = "); printf(ns, bestSol);

  return 1;
}
