/* This program solves the 0-1 Knapsack problem using Branch and Bound
 * For information about the Knapsack problem, see
 * "Knapsack Problems. Algorithms and Computer Implementations." by
 * Silvano Martello and Paolo Toth. John Wiley and Sons.
 * For information on Branch and Bound use the book of Horowitz and Sahni
 * "Fundamentals of Computers Algorithms" (Pitman)
 * This is approach focuses in the aspects of general
 * parallel branch and bound and does not in the concrete
 * features of knapsack problems.
 * Author: Casiano Rodriguez Leon. Email: casiano@ull.es
*/
#undef FORK95

#ifdef FORK95
#include <fork.h>
#include <io.h>
#else
#include <stdio.h>
#endif
#include <stdlib.h>

#define MAX 4096
#define MAXHEAPSIZE 16
#define STRIDE 1
#define  INFINITY 8192

typedef int number;
typedef int boolean;

typedef struct {
	int k, C, P;
} problemt;

#define ns "%d\n"

number N, M;
number w[MAX], p[MAX];
number bestSol = -INFINITY;

problemt problems[MAXHEAPSIZE];
number head = 0, tail = 0;

FILE *in;

void readKnap(void) {
  number i;

  if ((in = fopen("knap.dat","r+")) == NULL) 
    fprintf(stderr, "Cannot open input file.\n");
  fscanf(in,ns,&N);
  fscanf(in,ns,&M);
  printf("N = "); printf(ns,N);
  printf("M = "); printf(ns,M);
  for(i=0;i<N;i++) {
    fscanf(in,ns,&w[i]);
    fscanf(in,ns,&p[i]);
    if (i%STRIDE == 0) {
      printf("w = "); printf(ns,w[i]);
      printf("p = "); printf(ns,p[i]);
	 }
  }
  w[N] = INFINITY; p[N] = 0; /* sentinel object */
  fclose(in);
}

void lowerUpper(problemt Prob, number *L, number *U) {
  /* naive approach */
  number i, weig = 0, prof = Prob.P;

  if (Prob.C < 0) {
	 *L = -INFINITY; *U = -INFINITY;
  }
  else {
	 for(i=Prob.k; weig <= Prob.C; i++) {
		weig += w[i]; prof += p[i];
	 }
	 /* invariant:  weig > prob.C */
	 i--;
	 weig -= w[i];
	 prof -= p[i];
	 *L = prof;  /* see Martello and Toth pag 16. Th 2.1 */
	 *U = prof+(p[i]*(Prob.C-weig))/w[i];
  }
}

void setProblem(problemt *Prob, number k, number C, number P) {

  Prob->k = k;
  Prob->C = C;
  Prob->P = P;
}

void getProblem(problemt *Prob) {

  setProblem(Prob, problems[head].k, problems[head].C, problems[head].P);
  head = (head+1)% MAXHEAPSIZE;
}

void insertProblem(problemt Prob) {
  number next;

  next = (tail+1)%MAXHEAPSIZE;

  if (next == head) {
	 printf("\nError. The heap of problems is full.");
	 printf("\nIncrease the value of MAXHEAPSIZE. Current value is (%d)\n",MAXHEAPSIZE);
	 exit(1);
  }
  setProblem(&problems[tail],Prob.k, Prob.C, Prob.P);
  tail = next;
}

number numProblems(void) { /* number of problems in the queue */
  return (((tail-head)+MAXHEAPSIZE)%MAXHEAPSIZE);
}

number numChildren(problemt Prob) {
  return ((Prob.k < N) ? 2:0);
}

void child(problemt father,number order, problemt *child) {

  setProblem(child,
				 father.k+1,
				 father.C-(1-order)*w[father.k],
				 father.P+(1-order)*p[father.k]
				);
}

void branchAndBound(void) {
  problemt Prob, Probk;
  number nC, k, L, U;

  while (numProblems() > 0) {
	 getProblem(&Prob);
	 nC = numChildren(Prob);
	 if (nC != 0) {
		lowerUpper(Prob, &L, &U);
		if (bestSol < L) bestSol = L;	/* L <= bestSol <= U */
		if (bestSol < U) {
		  for(k=0; k<nC; k++) {
			child(Prob,k,&Probk);
			insertProblem(Probk);
		  }
		}
	 } /* end k < N */
  } /* end while */
}

int main(void) {
  problemt Prob;

  readKnap();

  setProblem(&Prob, 0, M, 0);
  insertProblem(Prob);
  branchAndBound();
  printf("\nsol = "); printf(ns, bestSol);

  return 1;
}
